import unittest

import numpy as np
import pandas as pd
import networkx as nx

from pgmpy.models.BayesianModel import BayesianModel
from pgmpy.estimators import TreeSearch
from pgmpy.factors.discrete import TabularCPD
from pgmpy.sampling import BayesianModelSampling


class TestTreeSearch(unittest.TestCase):
    def setUp(self):
        # test data for chow-liu
        self.data11 = pd.DataFrame(
            data=[[1, 1, 0], [1, 1, 1], [1, 1, 0]], columns=["B", "C", "D"]
        )

        # test data for chow-liu
        self.data12 = pd.DataFrame(
            np.random.randint(low=0, high=2, size=(100, 5)),
            columns=["A", "B", "C", "D", "E"],
        )

        # test data for chow-liu
        model = BayesianModel(
            [("A", "B"), ("A", "C"), ("B", "D"), ("B", "E"), ("C", "F")]
        )
        cpd_a = TabularCPD("A", 2, [[0.4], [0.6]])
        cpd_b = TabularCPD(
            "B",
            3,
            [[0.6, 0.2], [0.3, 0.5], [0.1, 0.3]],
            evidence=["A"],
            evidence_card=[2],
        )
        cpd_c = TabularCPD(
            "C", 2, [[0.3, 0.4], [0.7, 0.6]], evidence=["A"], evidence_card=[2]
        )
        cpd_d = TabularCPD(
            "D",
            3,
            [[0.5, 0.3, 0.1], [0.4, 0.4, 0.8], [0.1, 0.3, 0.1]],
            evidence=["B"],
            evidence_card=[3],
        )
        cpd_e = TabularCPD(
            "E",
            2,
            [[0.3, 0.5, 0.2], [0.7, 0.5, 0.8]],
            evidence=["B"],
            evidence_card=[3],
        )
        cpd_f = TabularCPD(
            "F",
            3,
            [[0.3, 0.6], [0.5, 0.2], [0.2, 0.2]],
            evidence=["C"],
            evidence_card=[2],
        )

        model.add_cpds(cpd_a, cpd_b, cpd_c, cpd_d, cpd_e, cpd_f)
        inference = BayesianModelSampling(model)
        self.data13 = inference.forward_sample(size=10000, return_type="dataframe")

        # test data for TAN
        self.data21 = pd.DataFrame(
            data=[[0, 1, 1, 0], [1, 1, 1, 1], [0, 1, 1, 0]],
            columns=["A", "B", "C", "D"],
        )

        # test data for TAN
        model = BayesianModel(
            [
                ("A", "R"),
                ("A", "B"),
                ("A", "C"),
                ("A", "D"),
                ("A", "E"),
                ("R", "B"),
                ("R", "C"),
                ("R", "D"),
                ("R", "E"),
            ]
        )
        cpd_a = TabularCPD("A", 2, [[0.7], [0.3]])
        cpd_r = TabularCPD(
            "R",
            3,
            [[0.6, 0.2], [0.3, 0.5], [0.1, 0.3]],
            evidence=["A"],
            evidence_card=[2],
        )
        cpd_b = TabularCPD(
            "B",
            3,
            [
                [0.1, 0.1, 0.2, 0.2, 0.7, 0.1],
                [0.1, 0.3, 0.1, 0.2, 0.1, 0.2],
                [0.8, 0.6, 0.7, 0.6, 0.2, 0.7],
            ],
            evidence=["A", "R"],
            evidence_card=[2, 3],
        )
        cpd_c = TabularCPD(
            "C",
            2,
            [[0.7, 0.2, 0.2, 0.5, 0.1, 0.3], [0.3, 0.8, 0.8, 0.5, 0.9, 0.7]],
            evidence=["A", "R"],
            evidence_card=[2, 3],
        )
        cpd_d = TabularCPD(
            "D",
            3,
            [
                [0.3, 0.8, 0.2, 0.8, 0.4, 0.7],
                [0.4, 0.1, 0.4, 0.1, 0.1, 0.1],
                [0.3, 0.1, 0.4, 0.1, 0.5, 0.2],
            ],
            evidence=["A", "R"],
            evidence_card=[2, 3],
        )
        cpd_e = TabularCPD(
            "E",
            2,
            [[0.5, 0.6, 0.6, 0.5, 0.5, 0.4], [0.5, 0.4, 0.4, 0.5, 0.5, 0.6]],
            evidence=["A", "R"],
            evidence_card=[2, 3],
        )
        model.add_cpds(cpd_a, cpd_r, cpd_b, cpd_c, cpd_d, cpd_e)
        inference = BayesianModelSampling(model)
        self.data22 = inference.forward_sample(size=10000, return_type="dataframe")

    def test_estimate_chow_liu(self):
        # learn tree structure using D as root node
        for n_jobs in [-1, 1]:
            est = TreeSearch(self.data11, root_node="D", n_jobs=n_jobs)
            dag = est.estimate(estimator_type="chow-liu")

            # check number of nodes and edges are as expected
            self.assertCountEqual(dag.nodes(), ["B", "C", "D"])
            self.assertCountEqual(dag.edges(), [("D", "B"), ("D", "C")])

            # check tree structure exists
            self.assertTrue(dag.has_edge("D", "B"))
            self.assertTrue(dag.has_edge("D", "C"))

            # learn tree structure using B as root node
            est = TreeSearch(self.data11, root_node="B", n_jobs=n_jobs)
            dag = est.estimate(estimator_type="chow-liu")

            # check number of nodes and edges are as expected
            self.assertCountEqual(dag.nodes(), ["B", "C", "D"])
            self.assertCountEqual(dag.edges(), [("B", "D"), ("D", "C")])

            # check tree structure exists
            self.assertTrue(dag.has_edge("B", "D"))
            self.assertTrue(dag.has_edge("D", "C"))

            # check invalid root node
            with self.assertRaises(ValueError):
                est = TreeSearch(self.data11, root_node="A", n_jobs=n_jobs)

            # learn graph structure
            est = TreeSearch(self.data12, root_node="A", n_jobs=n_jobs)
            dag = est.estimate(estimator_type="chow-liu")

            # check number of nodes and edges are as expected
            self.assertCountEqual(dag.nodes(), ["A", "B", "C", "D", "E"])
            self.assertTrue(nx.is_tree(dag))

            # learn tree structure using A as root node
            est = TreeSearch(self.data13, root_node="A", n_jobs=n_jobs)
            dag = est.estimate(estimator_type="chow-liu")

            # check number of nodes and edges are as expected
            self.assertCountEqual(dag.nodes(), ["A", "B", "C", "D", "E", "F"])
            self.assertCountEqual(
                dag.edges(),
                [("A", "B"), ("A", "C"), ("B", "D"), ("B", "E"), ("C", "F")],
            )

            # check tree structure exists
            self.assertTrue(dag.has_edge("A", "B"))
            self.assertTrue(dag.has_edge("A", "C"))
            self.assertTrue(dag.has_edge("B", "D"))
            self.assertTrue(dag.has_edge("B", "E"))
            self.assertTrue(dag.has_edge("C", "F"))

    def test_estimate_tan(self):
        for n_jobs in [-1, 1]:
            # learn graph structure
            est = TreeSearch(self.data21, root_node="D", n_jobs=n_jobs)
            dag = est.estimate(estimator_type="tan", class_node="A")

            # check number of nodes and edges are as expected
            self.assertCountEqual(dag.nodes(), ["A", "B", "C", "D"])
            self.assertCountEqual(
                dag.edges(),
                [("A", "B"), ("A", "C"), ("A", "D"), ("D", "B"), ("D", "C")],
            )

            # check directed edge between dependent and independent variables
            self.assertTrue(dag.has_edge("A", "B"))
            self.assertTrue(dag.has_edge("A", "C"))
            self.assertTrue(dag.has_edge("A", "D"))

            # check tree structure exists over independent variables
            self.assertTrue(dag.has_edge("D", "B"))
            self.assertTrue(dag.has_edge("D", "C"))

            # check invalid root node
            with self.assertRaises(ValueError):
                est = TreeSearch(self.data21, root_node="X", n_jobs=n_jobs)

            # check invalid class node
            est = TreeSearch(self.data21, root_node="D", n_jobs=n_jobs)
            with self.assertRaises(ValueError):
                est.estimate(estimator_type="tan", class_node="X")

            est = TreeSearch(self.data21, root_node="D", n_jobs=n_jobs)
            with self.assertRaises(ValueError):
                est.estimate(estimator_type="tan", class_node="D")

            # learn graph structure
            est = TreeSearch(self.data22, root_node="R", n_jobs=n_jobs)
            dag = est.estimate(estimator_type="tan", class_node="A")

            # check number of nodes and edges are as expected
            self.assertCountEqual(dag.nodes(), ["A", "B", "C", "D", "E", "R"])
            self.assertCountEqual(
                dag.edges(),
                [
                    ("A", "B"),
                    ("A", "C"),
                    ("A", "D"),
                    ("A", "E"),
                    ("A", "R"),
                    ("R", "B"),
                    ("R", "C"),
                    ("R", "D"),
                    ("R", "E"),
                ],
            )

            # check directed edge between class and independent variables
            self.assertTrue(dag.has_edge("A", "B"))
            self.assertTrue(dag.has_edge("A", "C"))
            self.assertTrue(dag.has_edge("A", "D"))
            self.assertTrue(dag.has_edge("A", "E"))

            # check tree structure exists over independent variables
            self.assertTrue(dag.has_edge("R", "B"))
            self.assertTrue(dag.has_edge("R", "C"))
            self.assertTrue(dag.has_edge("R", "D"))
            self.assertTrue(dag.has_edge("R", "E"))

    def tearDown(self):
        del self.data11
        del self.data12
        del self.data21
        del self.data22



import unittest

import networkx as nx
import pandas as pd
import numpy as np

from pgmpy.models import NaiveBayes
from pgmpy.independencies import Independencies


class TestBaseModelCreation(unittest.TestCase):
    def setUp(self):
        self.G = NaiveBayes()

    def test_class_init_without_data(self):
        self.assertIsInstance(self.G, nx.DiGraph)

    def test_class_init_with_data_string(self):
        self.g = NaiveBayes(feature_vars=["b", "c"], dependent_var="a")
        self.assertCountEqual(list(self.g.nodes()), ["a", "b", "c"])
        self.assertCountEqual(list(self.g.edges()), [("a", "b"), ("a", "c")])
        self.assertEqual(self.g.dependent, "a")
        self.assertSetEqual(self.g.features, {"b", "c"})

    def test_class_init_with_data_nonstring(self):
        self.g = NaiveBayes(feature_vars=[2, 3], dependent_var=1)
        self.assertCountEqual(list(self.g.nodes()), [1, 2, 3])
        self.assertCountEqual(list(self.g.edges()), [(1, 2), (1, 3)])
        self.assertEqual(self.g.dependent, 1)
        self.assertSetEqual(self.g.features, {2, 3})

    def test_add_node_string(self):
        self.G.add_node("a")
        self.assertListEqual(list(self.G.nodes()), ["a"])

    def test_add_node_nonstring(self):
        self.G.add_node(1)
        self.assertListEqual(list(self.G.nodes()), [1])

    def test_add_nodes_from_string(self):
        self.G.add_nodes_from(["a", "b", "c", "d"])
        self.assertCountEqual(list(self.G.nodes()), ["a", "b", "c", "d"])

    def test_add_nodes_from_non_string(self):
        self.G.add_nodes_from([1, 2, 3, 4])
        self.assertCountEqual(list(self.G.nodes()), [1, 2, 3, 4])

    def test_add_edge_string(self):
        self.G.add_edge("a", "b")
        self.assertCountEqual(list(self.G.nodes()), ["a", "b"])
        self.assertListEqual(list(self.G.edges()), [("a", "b")])
        self.assertEqual(self.G.dependent, "a")
        self.assertSetEqual(self.G.features, {"b"})

        self.G.add_nodes_from(["c", "d"])
        self.G.add_edge("a", "c")
        self.G.add_edge("a", "d")
        self.assertCountEqual(list(self.G.nodes()), ["a", "b", "c", "d"])
        self.assertCountEqual(
            list(self.G.edges()), [("a", "b"), ("a", "c"), ("a", "d")]
        )
        self.assertEqual(self.G.dependent, "a")
        self.assertSetEqual(self.G.features, {"b", "c", "d"})

        self.assertRaises(ValueError, self.G.add_edge, "b", "c")
        self.assertRaises(ValueError, self.G.add_edge, "d", "f")
        self.assertRaises(ValueError, self.G.add_edge, "e", "f")
        self.assertRaises(ValueError, self.G.add_edges_from, [("a", "e"), ("b", "f")])
        self.assertRaises(ValueError, self.G.add_edges_from, [("b", "f")])

    def test_add_edge_nonstring(self):
        self.G.add_edge(1, 2)
        self.assertCountEqual(list(self.G.nodes()), [1, 2])
        self.assertListEqual(list(self.G.edges()), [(1, 2)])
        self.assertEqual(self.G.dependent, 1)
        self.assertSetEqual(self.G.features, {2})

        self.G.add_nodes_from([3, 4])
        self.G.add_edge(1, 3)
        self.G.add_edge(1, 4)
        self.assertCountEqual(list(self.G.nodes()), [1, 2, 3, 4])
        self.assertCountEqual(list(self.G.edges()), [(1, 2), (1, 3), (1, 4)])
        self.assertEqual(self.G.dependent, 1)
        self.assertSetEqual(self.G.features, {2, 3, 4})

        self.assertRaises(ValueError, self.G.add_edge, 2, 3)
        self.assertRaises(ValueError, self.G.add_edge, 3, 6)
        self.assertRaises(ValueError, self.G.add_edge, 5, 6)
        self.assertRaises(ValueError, self.G.add_edges_from, [(1, 5), (2, 6)])
        self.assertRaises(ValueError, self.G.add_edges_from, [(2, 6)])

    def test_add_edge_selfloop(self):
        self.assertRaises(ValueError, self.G.add_edge, "a", "a")
        self.assertRaises(ValueError, self.G.add_edge, 1, 1)

    def test_add_edges_from_self_loop(self):
        self.assertRaises(ValueError, self.G.add_edges_from, [("a", "a")])

    def test_update_node_parents_bm_constructor(self):
        self.g = NaiveBayes(feature_vars=["b", "c"], dependent_var="a")
        self.assertListEqual(list(self.g.predecessors("a")), [])
        self.assertListEqual(list(self.g.predecessors("b")), ["a"])
        self.assertListEqual(list(self.g.predecessors("c")), ["a"])

    def test_update_node_parents(self):
        self.G.add_nodes_from(["a", "b", "c"])
        self.G.add_edges_from([("a", "b"), ("a", "c")])
        self.assertListEqual(list(self.G.predecessors("a")), [])
        self.assertListEqual(list(self.G.predecessors("b")), ["a"])
        self.assertListEqual(list(self.G.predecessors("c")), ["a"])

    def tearDown(self):
        del self.G


class TestNaiveBayesMethods(unittest.TestCase):
    def setUp(self):
        self.G1 = NaiveBayes(feature_vars=["b", "c", "d", "e"], dependent_var="a")
        self.G2 = NaiveBayes(feature_vars=["g", "l", "s"], dependent_var="d")

    def test_local_independencies(self):
        self.assertEqual(self.G1.local_independencies("a"), Independencies())
        self.assertEqual(
            self.G1.local_independencies("b"),
            Independencies(["b", ["e", "c", "d"], "a"]),
        )
        self.assertEqual(
            self.G1.local_independencies("c"),
            Independencies(["c", ["e", "b", "d"], "a"]),
        )
        self.assertEqual(
            self.G1.local_independencies("d"),
            Independencies(["d", ["b", "c", "e"], "a"]),
        )

    def test_active_trail_nodes(self):
        self.assertListEqual(
            sorted(self.G2.active_trail_nodes("d")), ["d", "g", "l", "s"]
        )
        self.assertListEqual(
            sorted(self.G2.active_trail_nodes("g")), ["d", "g", "l", "s"]
        )
        self.assertListEqual(
            sorted(self.G2.active_trail_nodes("l")), ["d", "g", "l", "s"]
        )
        self.assertListEqual(
            sorted(self.G2.active_trail_nodes("s")), ["d", "g", "l", "s"]
        )

    def test_active_trail_nodes_args(self):
        self.assertListEqual(
            sorted(self.G2.active_trail_nodes("d", observed="g")), ["d", "l", "s"]
        )
        self.assertListEqual(
            sorted(self.G2.active_trail_nodes("l", observed="g")), ["d", "l", "s"]
        )
        self.assertListEqual(
            sorted(self.G2.active_trail_nodes("s", observed=["g", "l"])), ["d", "s"]
        )
        self.assertListEqual(
            sorted(self.G2.active_trail_nodes("s", observed=["d", "l"])), ["s"]
        )

    def test_get_ancestors_of(self):
        self.assertListEqual(sorted(self.G1._get_ancestors_of("b")), ["a", "b"])
        self.assertListEqual(sorted(self.G1._get_ancestors_of("e")), ["a", "e"])
        self.assertListEqual(sorted(self.G1._get_ancestors_of("a")), ["a"])
        self.assertListEqual(
            sorted(self.G1._get_ancestors_of(["b", "e"])), ["a", "b", "e"]
        )

    def tearDown(self):
        del self.G1
        del self.G2


class TestNaiveBayesFit(unittest.TestCase):
    def setUp(self):
        self.model1 = NaiveBayes()
        self.model2 = NaiveBayes(feature_vars=["B"], dependent_var="A")

    def test_fit_model_creation(self):
        values = pd.DataFrame(
            np.random.randint(low=0, high=2, size=(1000, 5)),
            columns=["A", "B", "C", "D", "E"],
        )

        self.model1.fit(values, "A")
        self.assertCountEqual(self.model1.nodes(), ["A", "B", "C", "D", "E"])
        self.assertCountEqual(
            self.model1.edges(), [("A", "B"), ("A", "C"), ("A", "D"), ("A", "E")]
        )
        self.assertEqual(self.model1.dependent, "A")
        self.assertSetEqual(self.model1.features, {"B", "C", "D", "E"})

        self.model2.fit(values)
        self.assertCountEqual(self.model1.nodes(), ["A", "B", "C", "D", "E"])
        self.assertCountEqual(
            self.model1.edges(), [("A", "B"), ("A", "C"), ("A", "D"), ("A", "E")]
        )
        self.assertEqual(self.model2.dependent, "A")
        self.assertSetEqual(self.model2.features, {"B", "C", "D", "E"})

    def test_fit_model_creation_exception(self):
        values = pd.DataFrame(
            np.random.randint(low=0, high=2, size=(1000, 5)),
            columns=["A", "B", "C", "D", "E"],
        )
        values2 = pd.DataFrame(
            np.random.randint(low=0, high=2, size=(1000, 3)), columns=["C", "D", "E"]
        )

        self.assertRaises(ValueError, self.model1.fit, values)
        self.assertRaises(ValueError, self.model1.fit, values2)
        self.assertRaises(ValueError, self.model2.fit, values2, "A")

    def tearDown(self):
        del self.model1
        del self.model2



#!/usr/bin/env python3

import unittest

from pgmpy.base import DAG, PDAG
import pgmpy.tests.help_functions as hf
import networkx as nx


class TestDAGCreation(unittest.TestCase):
    def setUp(self):
        self.graph = DAG()

    def test_class_init_without_data(self):
        self.assertIsInstance(self.graph, DAG)

    def test_class_init_with_data_string(self):
        self.graph = DAG([("a", "b"), ("b", "c")])
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c"])
        self.assertListEqual(
            hf.recursive_sorted(self.graph.edges()), [["a", "b"], ["b", "c"]]
        )

    def test_add_node_string(self):
        self.graph.add_node("a")
        self.assertListEqual(list(self.graph.nodes()), ["a"])

    def test_add_node_nonstring(self):
        self.graph.add_node(1)

    def test_add_nodes_from_string(self):
        self.graph.add_nodes_from(["a", "b", "c", "d"])
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c", "d"])

    def test_add_nodes_from_non_string(self):
        self.graph.add_nodes_from([1, 2, 3, 4])

    def test_add_node_weight(self):
        self.graph.add_node("weighted_a", 0.3)
        self.assertEqual(self.graph.nodes["weighted_a"]["weight"], 0.3)

    def test_add_nodes_from_weight(self):
        self.graph.add_nodes_from(["weighted_b", "weighted_c"], [0.5, 0.6])
        self.assertEqual(self.graph.nodes["weighted_b"]["weight"], 0.5)
        self.assertEqual(self.graph.nodes["weighted_c"]["weight"], 0.6)

        self.graph.add_nodes_from(["e", "f"])
        self.assertEqual(self.graph.nodes["e"]["weight"], None)
        self.assertEqual(self.graph.nodes["f"]["weight"], None)

    def test_add_edge_string(self):
        self.graph.add_edge("d", "e")
        self.assertListEqual(sorted(self.graph.nodes()), ["d", "e"])
        self.assertListEqual(list(self.graph.edges()), [("d", "e")])
        self.graph.add_nodes_from(["a", "b", "c"])
        self.graph.add_edge("a", "b")
        self.assertListEqual(
            hf.recursive_sorted(self.graph.edges()), [["a", "b"], ["d", "e"]]
        )

    def test_add_edge_nonstring(self):
        self.graph.add_edge(1, 2)

    def test_add_edges_from_string(self):
        self.graph.add_edges_from([("a", "b"), ("b", "c")])
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c"])
        self.assertListEqual(
            hf.recursive_sorted(self.graph.edges()), [["a", "b"], ["b", "c"]]
        )
        self.graph.add_nodes_from(["d", "e", "f"])
        self.graph.add_edges_from([("d", "e"), ("e", "f")])
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c", "d", "e", "f"])
        self.assertListEqual(
            hf.recursive_sorted(self.graph.edges()),
            hf.recursive_sorted([("a", "b"), ("b", "c"), ("d", "e"), ("e", "f")]),
        )

    def test_add_edges_from_nonstring(self):
        self.graph.add_edges_from([(1, 2), (2, 3)])

    def test_add_edge_weight(self):
        self.graph.add_edge("a", "b", weight=0.3)
        if nx.__version__.startswith("1"):
            self.assertEqual(self.graph.edge["a"]["b"]["weight"], 0.3)
        else:
            self.assertEqual(self.graph.adj["a"]["b"]["weight"], 0.3)

    def test_add_edges_from_weight(self):
        self.graph.add_edges_from([("b", "c"), ("c", "d")], weights=[0.5, 0.6])
        if nx.__version__.startswith("1"):
            self.assertEqual(self.graph.edge["b"]["c"]["weight"], 0.5)
            self.assertEqual(self.graph.edge["c"]["d"]["weight"], 0.6)

            self.graph.add_edges_from([("e", "f")])
            self.assertEqual(self.graph.edge["e"]["f"]["weight"], None)
        else:
            self.assertEqual(self.graph.adj["b"]["c"]["weight"], 0.5)
            self.assertEqual(self.graph.adj["c"]["d"]["weight"], 0.6)

            self.graph.add_edges_from([("e", "f")])
            self.assertEqual(self.graph.adj["e"]["f"]["weight"], None)

    def test_update_node_parents_bm_constructor(self):
        self.graph = DAG([("a", "b"), ("b", "c")])
        self.assertListEqual(list(self.graph.predecessors("a")), [])
        self.assertListEqual(list(self.graph.predecessors("b")), ["a"])
        self.assertListEqual(list(self.graph.predecessors("c")), ["b"])

    def test_update_node_parents(self):
        self.graph.add_nodes_from(["a", "b", "c"])
        self.graph.add_edges_from([("a", "b"), ("b", "c")])
        self.assertListEqual(list(self.graph.predecessors("a")), [])
        self.assertListEqual(list(self.graph.predecessors("b")), ["a"])
        self.assertListEqual(list(self.graph.predecessors("c")), ["b"])

    def test_get_leaves(self):
        self.graph.add_edges_from(
            [("A", "B"), ("B", "C"), ("B", "D"), ("D", "E"), ("D", "F"), ("A", "G")]
        )
        self.assertEqual(sorted(self.graph.get_leaves()), sorted(["C", "G", "E", "F"]))

    def test_get_roots(self):
        self.graph.add_edges_from(
            [("A", "B"), ("B", "C"), ("B", "D"), ("D", "E"), ("D", "F"), ("A", "G")]
        )
        self.assertEqual(["A"], self.graph.get_roots())
        self.graph.add_edge("H", "G")
        self.assertEqual(sorted(["A", "H"]), sorted(self.graph.get_roots()))

    def test_init_with_cycle(self):
        self.assertRaises(ValueError, DAG, [("a", "a")])
        self.assertRaises(ValueError, DAG, [("a", "b"), ("b", "a")])
        self.assertRaises(ValueError, DAG, [("a", "b"), ("b", "c"), ("c", "a")])

    def test_get_ancestral_graph(self):
        dag = DAG([("A", "C"), ("B", "C"), ("D", "A"), ("D", "B")])
        anc_dag = dag.get_ancestral_graph(["A", "B"])
        self.assertEqual(set(anc_dag.edges()), set([("D", "A"), ("D", "B")]))
        self.assertRaises(ValueError, dag.get_ancestral_graph, ["A", "gibber"])

    def tearDown(self):
        del self.graph


class TestDAGMoralization(unittest.TestCase):
    def setUp(self):
        self.graph = DAG()
        self.graph.add_edges_from([("diff", "grade"), ("intel", "grade")])

    def test_get_parents(self):
        self.assertListEqual(sorted(self.graph.get_parents("grade")), ["diff", "intel"])

    def test_moralize(self):
        moral_graph = self.graph.moralize()
        self.assertListEqual(
            hf.recursive_sorted(moral_graph.edges()),
            [["diff", "grade"], ["diff", "intel"], ["grade", "intel"]],
        )

    def test_moralize_disconnected(self):
        graph_copy = self.graph.copy()
        graph_copy.add_node("disconnected")
        moral_graph = graph_copy.moralize()
        self.assertListEqual(
            hf.recursive_sorted(moral_graph.edges()),
            [["diff", "grade"], ["diff", "intel"], ["grade", "intel"]],
        )
        self.assertEqual(
            sorted(moral_graph.nodes()), ["diff", "disconnected", "grade", "intel"]
        )

    def test_get_children(self):
        self.assertListEqual(sorted(self.graph.get_children("diff")), ["grade"])

    def tearDown(self):
        del self.graph


class TestDoOperator(unittest.TestCase):
    def setUp(self):
        self.graph = DAG()
        self.graph.add_edges_from([("X", "A"), ("A", "Y"), ("A", "B")])

    def test_do(self):
        dag_do_x = self.graph.do("A")
        self.assertEqual(set(dag_do_x.nodes()), set(self.graph.nodes()))
        self.assertEqual(sorted(list(dag_do_x.edges())), [("A", "B"), ("A", "Y")])


class TestPDAG(unittest.TestCase):
    def setUp(self):
        self.pdag_mix = PDAG(
            directed_ebunch=[("A", "C"), ("D", "C")],
            undirected_ebunch=[("B", "A"), ("B", "D")],
        )
        self.pdag_dir = PDAG(
            directed_ebunch=[("A", "B"), ("D", "B"), ("A", "C"), ("D", "C")]
        )
        self.pdag_undir = PDAG(
            undirected_ebunch=[("A", "C"), ("D", "C"), ("B", "A"), ("B", "D")]
        )

    def test_init_normal(self):
        # Mix directed and undirected
        directed_edges = [("A", "C"), ("D", "C")]
        undirected_edges = [("B", "A"), ("B", "D")]
        pdag = PDAG(directed_ebunch=directed_edges, undirected_ebunch=undirected_edges)
        expected_edges = {
            ("A", "C"),
            ("D", "C"),
            ("A", "B"),
            ("B", "A"),
            ("B", "D"),
            ("D", "B"),
        }
        self.assertEqual(set(pdag.edges()), expected_edges)
        self.assertEqual(set(pdag.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag.directed_edges, set(directed_edges))
        self.assertEqual(pdag.undirected_edges, set(undirected_edges))

        # Only undirected
        undirected_edges = [("A", "C"), ("D", "C"), ("B", "A"), ("B", "D")]
        pdag = PDAG(undirected_ebunch=undirected_edges)
        expected_edges = {
            ("A", "C"),
            ("C", "A"),
            ("D", "C"),
            ("C", "D"),
            ("B", "A"),
            ("A", "B"),
            ("B", "D"),
            ("D", "B"),
        }
        self.assertEqual(set(pdag.edges()), expected_edges)
        self.assertEqual(set(pdag.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag.directed_edges, set())
        self.assertEqual(pdag.undirected_edges, set(undirected_edges))

        # Only directed
        directed_edges = [("A", "B"), ("D", "B"), ("A", "C"), ("D", "C")]
        pdag = PDAG(directed_ebunch=directed_edges)
        self.assertEqual(set(pdag.edges()), set(directed_edges))
        self.assertEqual(set(pdag.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag.directed_edges, set(directed_edges))
        self.assertEqual(pdag.undirected_edges, set())

        # TODO: Fix the cycle issue.
        # Test cycle
        # directed_edges = [('A', 'C')]
        # undirected_edges = [('A', 'B'), ('B', 'D'), ('D', 'C')]
        # self.assertRaises(ValueError, PDAG, directed_ebunch=directed_edges, undirected_ebunch=undirected_edges)

    def test_copy(self):
        pdag_copy = self.pdag_mix.copy()
        expected_edges = {
            ("A", "C"),
            ("D", "C"),
            ("A", "B"),
            ("B", "A"),
            ("B", "D"),
            ("D", "B"),
        }
        expected_dir = [("A", "C"), ("D", "C")]
        expected_undir = [("B", "A"), ("B", "D")]
        self.assertEqual(set(pdag_copy.edges()), expected_edges)
        self.assertEqual(set(pdag_copy.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag_copy.directed_edges, set([("A", "C"), ("D", "C")]))
        self.assertEqual(pdag_copy.undirected_edges, set([("B", "A"), ("B", "D")]))

    def test_pdag_to_dag(self):
        # PDAG no: 1  Possibility of creating a v-structure
        pdag = PDAG(
            directed_ebunch=[("A", "B"), ("C", "B")],
            undirected_ebunch=[("C", "D"), ("D", "A")],
        )
        dag = pdag.to_dag()
        self.assertTrue(("A", "B") in dag.edges())
        self.assertTrue(("C", "B") in dag.edges())
        self.assertFalse((("A", "D") in dag.edges()) and (("C", "D") in dag.edges()))
        self.assertTrue(len(dag.edges()) == 4)

        # PDAG no: 2  No possibility of creation of v-structure.
        pdag = PDAG(
            directed_ebunch=[("B", "C"), ("A", "C")], undirected_ebunch=[("A", "D")]
        )
        dag = pdag.to_dag()
        self.assertTrue(("B", "C") in dag.edges())
        self.assertTrue(("A", "C") in dag.edges())
        self.assertTrue((("A", "D") in dag.edges()) or (("D", "A") in dag.edges()))

        # PDAG no: 3  Already existing v-structure, possiblity to add another
        pdag = PDAG(
            directed_ebunch=[("B", "C"), ("A", "C")], undirected_ebunch=[("C", "D")]
        )
        dag = pdag.to_dag()
        expected_edges = {("B", "C"), ("C", "D"), ("A", "C")}
        self.assertEqual(expected_edges, set(dag.edges()))



#!/usr/bin/env python3
import sys
import unittest
import numpy as np
from pandas import DataFrame
from mock import patch, call


from pgmpy.factors.discrete import State
from pgmpy.models import MarkovChain as MC


class TestMarkovChain(unittest.TestCase):
    def setUp(self):
        self.variables = ["intel", "diff", "grade"]
        self.card = [3, 2, 3]
        self.cardinalities = {"intel": 3, "diff": 2, "grade": 3}
        self.intel_tm = {
            0: {0: 0.1, 1: 0.25, 2: 0.65},
            1: {0: 0.5, 1: 0.3, 2: 0.2},
            2: {0: 0.3, 1: 0.3, 2: 0.4},
        }
        self.intel_tm_matrix = np.array(
            [[0.1, 0.25, 0.65], [0.5, 0.3, 0.2], [0.3, 0.3, 0.4]]
        )
        self.diff_tm = {0: {0: 0.3, 1: 0.7}, 1: {0: 0.75, 1: 0.25}}
        self.diff_tm_matrix = np.array([[0.3, 0.7], [0.75, 0.25]])
        self.grade_tm = {
            0: {0: 0.4, 1: 0.2, 2: 0.4},
            1: {0: 0.9, 1: 0.05, 2: 0.05},
            2: {0: 0.1, 1: 0.4, 2: 0.5},
        }
        self.grade_tm_matrix = [[0.4, 0.2, 0.4], [0.9, 0.05, 0.05], [0.1, 0.4, 0.5]]
        self.start_state = [State("intel", 0), State("diff", 1), State("grade", 2)]
        self.model = MC()

        self.sample = DataFrame(index=range(200), columns=["a", "b"])
        self.sample.a = [1] * 100 + [0] * 100
        self.sample.b = [0] * 100 + [1] * 100

    def tearDown(self):
        del self.variables
        del self.card
        del self.cardinalities
        del self.intel_tm
        del self.diff_tm
        del self.grade_tm
        del self.start_state
        del self.model
        del self.sample

    @patch("pgmpy.models.MarkovChain._check_state", autospec=True)
    def test_init(self, check_state):
        model = MC(self.variables, self.card, self.start_state)
        self.assertListEqual(model.variables, self.variables)
        self.assertDictEqual(model.cardinalities, self.cardinalities)
        self.assertDictEqual(
            model.transition_models, {var: {} for var in self.variables}
        )
        check_state.assert_called_once_with(model, self.start_state)
        self.assertListEqual(model.state, self.start_state)

    def test_init_bad_variables_type(self):
        # variables is non-iterable
        self.assertRaises(ValueError, MC, variables=123)
        # variables is a string
        self.assertRaises(ValueError, MC, variables="abc")

    def test_init_bad_card_type(self):
        # card is non-iterable
        self.assertRaises(ValueError, MC, card=123)
        # card is a string
        self.assertRaises(ValueError, MC, card="abc")

    def test_init_less_args(self):
        model = MC()
        self.assertListEqual(model.variables, [])
        self.assertDictEqual(model.cardinalities, {})
        self.assertDictEqual(model.transition_models, {})
        self.assertIsNone(model.state)

    @patch("pgmpy.models.MarkovChain._check_state", autospec=True)
    def test_set_start_state_list(self, check_state):
        model = MC(["b", "a"], [1, 2])
        check_state.return_value = True
        model.set_start_state([State("a", 0), State("b", 1)])
        model_state = [State("b", 1), State("a", 0)]
        check_state.assert_called_once_with(model, model_state)
        self.assertEqual(model.state, model_state)

    def test_set_start_state_none(self):
        model = MC()
        model.state = "state"
        model.set_start_state(None)
        self.assertIsNone(model.state)

    def test_check_state_bad_type(self):
        model = MC()
        # state is non-iterable
        self.assertRaises(ValueError, model._check_state, 123)
        # state is a string
        self.assertRaises(ValueError, model._check_state, "abc")

    def test_check_state_bad_vars(self):
        model = MC()
        # state_vars and model_vars differ
        self.assertRaises(ValueError, model._check_state, [State(1, 2)])

    def test_check_state_bad_var_value(self):
        model = MC(["a"], [2])
        # value of variable >= cardinaliity
        self.assertRaises(ValueError, model._check_state, [State("a", 3)])

    def test_check_state_success(self):
        model = MC(["a"], [2])
        self.assertTrue(model._check_state([State("a", 1)]))

    def test_add_variable_new(self):
        model = MC(["a"], [2])
        model.add_variable("p", 3)
        self.assertIn("p", model.variables)
        self.assertEqual(model.cardinalities["p"], 3)
        self.assertDictEqual(model.transition_models["p"], {})

    def test_copy(self):
        model = MC(["a", "b"], [2, 2], [State("a", 0), State("b", 1)])
        model.add_transition_model("a", {0: {0: 0.1, 1: 0.9}, 1: {0: 0.2, 1: 0.8}})
        model.add_transition_model("b", {0: {0: 0.3, 1: 0.7}, 1: {0: 0.4, 1: 0.6}})
        copy = model.copy()

        self.assertIsInstance(copy, MC)
        self.assertEqual(sorted(model.variables), sorted(copy.variables))
        self.assertEqual(model.cardinalities, copy.cardinalities)
        self.assertEqual(model.transition_models, copy.transition_models)
        self.assertEqual(model.state, copy.state)

        model.add_variable("p", 1)
        model.set_start_state([State("a", 0), State("b", 1), State("p", 0)])
        model.add_transition_model("p", {0: {0: 1}})

        self.assertNotEqual(sorted(model.variables), sorted(copy.variables))
        self.assertEqual(sorted(["a", "b"]), sorted(copy.variables))
        self.assertNotEqual(model.cardinalities, copy.cardinalities)
        self.assertEqual({"a": 2, "b": 2}, copy.cardinalities)
        self.assertNotEqual(model.state, copy.state)
        self.assertEqual([State("a", 0), State("b", 1)], copy.state)
        self.assertNotEqual(model.transition_models, copy.transition_models)
        self.assertEqual(len(copy.transition_models), 2)
        self.assertEqual(
            copy.transition_models["a"], {0: {0: 0.1, 1: 0.9}, 1: {0: 0.2, 1: 0.8}}
        )
        self.assertEqual(
            copy.transition_models["b"], {0: {0: 0.3, 1: 0.7}, 1: {0: 0.4, 1: 0.6}}
        )

    @patch.object(sys.modules["pgmpy.models.MarkovChain"], "warn")
    def test_add_variable_existing(self, warn):
        model = MC(["p"], [2])
        model.add_variable("p", 3)
        self.assertEqual(warn.call_count, 1)

    @patch("pgmpy.models.MarkovChain.add_variable", autospec=True)
    def test_add_variables_from(self, add_var):
        model = MC()
        model.add_variables_from(self.variables, self.card)
        calls = [call(model, *p) for p in zip(self.variables, self.card)]
        add_var.assert_has_calls(calls)

    def test_add_transition_model_bad_type(self):
        model = MC()
        grade_tm_matrix_bad = [[0.1, 0.5, 0.4], [0.2, 0.2, 0.6], "abc"]
        # if transition_model is not a dict or np.array
        self.assertRaises(ValueError, model.add_transition_model, "var", 123)
        self.assertRaises(
            ValueError, model.add_transition_model, "var", grade_tm_matrix_bad
        )

    def test_add_transition_model_bad_states(self):
        model = MC(["var"], [2])
        # transition for state=1 not defined
        transition_model = {0: {0: 0.1, 1: 0.9}}
        self.assertRaises(
            ValueError, model.add_transition_model, "var", transition_model
        )

    def test_add_transition_model_bad_transition(self):
        model = MC(["var"], [2])
        # transition for state=1 is not a dict
        transition_model = {0: {0: 0.1, 1: 0.9}, 1: "abc"}
        self.assertRaises(
            ValueError, model.add_transition_model, "var", transition_model
        )

    def test_add_transition_model_bad_probability(self):
        model = MC(["var"], [2])
        transition_model = {0: {0: -0.1, 1: 1.1}, 1: {0: 0.5, 1: 0.5}}
        self.assertRaises(
            ValueError, model.add_transition_model, "var", transition_model
        )

    def test_add_transition_model_bad_probability_sum(self):
        model = MC(["var"], [2])
        # transition probabilities from state=0 do not sum to 1.0
        transition_model = {0: {0: 0.1, 1: 0.2}, 1: {0: 0.5, 1: 0.5}}
        self.assertRaises(
            ValueError, model.add_transition_model, "var", transition_model
        )

    def test_add_transition_model_success(self):
        model = MC(["var"], [2])
        transition_model = {0: {0: 0.3, 1: 0.7}, 1: {0: 0.5, 1: 0.5}}
        model.add_transition_model("var", transition_model)
        self.assertDictEqual(model.transition_models["var"], transition_model)

    def test_transition_model_bad_matrix_dimension(self):
        model = MC(["var"], [2])
        transition_model = np.array([0.3, 0.7])
        # check for square dimension of the matrix
        self.assertRaises(
            ValueError, model.add_transition_model, "var", transition_model
        )
        transition_model = np.array([[0.3, 0.6, 0.1], [0.3, 0.3, 0.4]])
        self.assertRaises(
            ValueError, model.add_transition_model, "var", transition_model
        )

    def test_transition_model_dict_to_matrix(self):
        model = MC(["var"], [2])
        transition_model = {0: {0: 0.3, 1: 0.7}, 1: {0: 0.5, 1: 0.5}}
        transition_model_matrix = np.array([[0.3, 0.7], [0.5, 0.5]])
        model.add_transition_model("var", transition_model_matrix)
        self.assertDictEqual(model.transition_models["var"], transition_model)

    def test_sample(self):
        model = MC(["a", "b"], [2, 2])
        model.transition_models["a"] = {0: {0: 0.1, 1: 0.9}, 1: {0: 0.2, 1: 0.8}}
        model.transition_models["b"] = {0: {0: 0.3, 1: 0.7}, 1: {0: 0.4, 1: 0.6}}
        sample = model.sample(start_state=[State("a", 0), State("b", 1)], size=2)
        self.assertEqual(len(sample), 2)
        self.assertEqual(list(sample.columns), ["a", "b"])
        self.assertTrue(list(sample.loc[0]) in [[0, 0], [0, 1], [1, 0], [1, 1]])
        self.assertTrue(list(sample.loc[1]) in [[0, 0], [0, 1], [1, 0], [1, 1]])

    @patch("pgmpy.models.MarkovChain.random_state", autospec=True)
    def test_sample_less_arg(self, random_state):
        model = MC(["a", "b"], [2, 2])
        random_state.return_value = [State("a", 0), State("b", 1)]
        sample = model.sample(size=1)
        random_state.assert_called_once_with(model)
        self.assertEqual(model.state, random_state.return_value)
        self.assertEqual(len(sample), 1)
        self.assertEqual(list(sample.columns), ["a", "b"])
        self.assertEqual(list(sample.loc[0]), [0, 1])

    @patch("pgmpy.models.MarkovChain.sample", autospec=True)
    def test_prob_from_sample(self, sample):
        model = MC(["a", "b"], [2, 2])
        sample.return_value = self.sample
        probabilites = model.prob_from_sample([State("a", 1), State("b", 0)])
        self.assertEqual(list(probabilites), [1] * 50 + [0] * 50)

    def test_is_stationarity_success(self):
        model = MC(["intel", "diff"], [2, 3])
        model.set_start_state([State("intel", 0), State("diff", 2)])
        intel_tm = {0: {0: 0.25, 1: 0.75}, 1: {0: 0.5, 1: 0.5}}
        model.add_transition_model("intel", intel_tm)
        diff_tm = {
            0: {0: 0.1, 1: 0.5, 2: 0.4},
            1: {0: 0.2, 1: 0.2, 2: 0.6},
            2: {0: 0.7, 1: 0.15, 2: 0.15},
        }
        model.add_transition_model("diff", diff_tm)
        self.assertTrue(model.is_stationarity)

    def test_is_stationarity_failure(self):
        model = MC(["intel", "diff"], [2, 3])
        model.set_start_state([State("intel", 0), State("diff", 2)])
        intel_tm = {0: {0: 0.25, 1: 0.75}, 1: {0: 0.5, 1: 0.5}}
        model.add_transition_model("intel", intel_tm)
        diff_tm = {
            0: {0: 0.1, 1: 0.5, 2: 0.4},
            1: {0: 0.2, 1: 0.2, 2: 0.6},
            2: {0: 0.7, 1: 0.15, 2: 0.15},
        }
        model.add_transition_model("diff", diff_tm)
        self.assertFalse(model.is_stationarity(0.002, None))

    @patch.object(sys.modules["pgmpy.models.MarkovChain"], "sample_discrete")
    def test_generate_sample(self, sample_discrete):
        model = MC(["a", "b"], [2, 2])
        model.transition_models["a"] = {0: {0: 0.1, 1: 0.9}, 1: {0: 0.2, 1: 0.8}}
        model.transition_models["b"] = {0: {0: 0.3, 1: 0.7}, 1: {0: 0.4, 1: 0.6}}
        sample_discrete.side_effect = [[1], [0]] * 2
        gen = model.generate_sample(start_state=[State("a", 0), State("b", 1)], size=2)
        samples = [sample for sample in gen]
        expected_samples = [[State("a", 1), State("b", 0)]] * 2
        self.assertEqual(samples, expected_samples)

    @patch.object(sys.modules["pgmpy.models.MarkovChain"], "sample_discrete")
    @patch("pgmpy.models.MarkovChain.random_state", autospec=True)
    def test_generate_sample_less_arg(self, random_state, sample_discrete):
        model = MC(["a", "b"], [2, 2])
        model.transition_models["a"] = {0: {0: 0.1, 1: 0.9}, 1: {0: 0.2, 1: 0.8}}
        model.transition_models["b"] = {0: {0: 0.3, 1: 0.7}, 1: {0: 0.4, 1: 0.6}}
        random_state.return_value = [State("a", 0), State("b", 1)]
        sample_discrete.side_effect = [[1], [0]] * 2
        gen = model.generate_sample(size=2)
        samples = [sample for sample in gen]
        expected_samples = [[State("a", 1), State("b", 0)]] * 2
        self.assertEqual(samples, expected_samples)

    def test_random_state(self):
        model = MC(["a", "b"], [2, 3])
        state = model.random_state()
        vars = [v for v, s in state]
        self.assertEqual(vars, ["a", "b"])
        self.assertGreaterEqual(state[0].state, 0)
        self.assertGreaterEqual(state[1].state, 0)
        self.assertLessEqual(state[0].state, 1)
        self.assertLessEqual(state[1].state, 2)



import unittest

from pandas import DataFrame
from numpy import NaN

from pgmpy.models import BayesianModel
from pgmpy.estimators import ParameterEstimator


class TestParameterEstimator(unittest.TestCase):
    def setUp(self):
        self.m1 = BayesianModel([("A", "C"), ("B", "C"), ("D", "B")])
        self.d1 = DataFrame(
            data={"A": [0, 0, 1], "B": [0, 1, 0], "C": [1, 1, 0], "D": ["X", "Y", "Z"]}
        )
        self.d2 = DataFrame(
            data={
                "A": [0, NaN, 1],
                "B": [0, 1, 0],
                "C": [1, 1, NaN],
                "D": [NaN, "Y", NaN],
            }
        )

    def test_state_count(self):
        e = ParameterEstimator(self.m1, self.d1)
        self.assertEqual(e.state_counts("A").values.tolist(), [[2], [1]])
        self.assertEqual(
            e.state_counts("C").values.tolist(),
            [[0.0, 0.0, 1.0, 0.0], [1.0, 1.0, 0.0, 0.0]],
        )

    def test_missing_data(self):
        e = ParameterEstimator(
            self.m1, self.d2, state_names={"C": [0, 1]}, complete_samples_only=False
        )
        self.assertEqual(
            e.state_counts("A", complete_samples_only=True).values.tolist(), [[0], [0]]
        )
        self.assertEqual(e.state_counts("A").values.tolist(), [[1], [1]])
        self.assertEqual(
            e.state_counts("C", complete_samples_only=True).values.tolist(),
            [[0, 0, 0, 0], [0, 0, 0, 0]],
        )
        self.assertEqual(
            e.state_counts("C").values.tolist(), [[0, 0, 0, 0], [1, 0, 0, 0]]
        )

    def tearDown(self):
        del self.m1
        del self.d1
